﻿# -*- coding: utf-8 -*-

import maya.api.OpenMaya as om2
import maya.cmds as mc

__DEBUG = False

#### ========================================================================
def __FnVMul(v0,v1):
	return([v0[0]*v1[0], v0[1]*v1[1], v0[2]*v1[2]])

def __FnVMulSub(v0,v1,v2):
	return([(v0[0]*v1[0])-v2[0], (v0[1]*v1[1])-v2[1], (v0[2]*v1[2])-v2[2]])

#### ========================================================================
def __ConvertDagPath_2To1(dagpMesh):
	sl = om.MSelectionList()
	sl.add(dagpMesh.fullPathName())
	dagp = om.MDagPath()
	sl.getDagPath(0, dagp)
	return(dagp)

#### ========================================================================
def __DetermineEdgeWeight(lyr):
	"""Modoは0~1(0.2が2に相当), LWは0~1(1が2に相当とする)
	maya(smooth/smoothPrev)での実効値0~2. 2以上は要mrSubdivApprox."""
	# LWOの最終更新がMODOならVMEDがある.優先.
	if(lyr.hasEdgeM): return(10.0, lyr.edgeWgtsM)

	# LWOの最終更新がLWだとVMEDは無い.
	# ModoでVMEDを消して,VMADが残っていた場合もこっちになるので要オプション？
	return(2.0, lyr.edgeWgts)

#### ----------------
def __ApplyCreaseEdge2(moMesh, lyr):
	"""エッジウェイト適用"""
	if((not lyr.hasEdge) and (not lyr.hasEdgeM)): return

	(coef, edgDict) = __DetermineEdgeWeight(lyr)

	FnMe = om2.MFnMesh(moMesh)
	idxs = om2.MUintArray()		# edgeIdx
	wgts = om2.MDoubleArray()	# wgt
	idxAPD = idxs.append
	wgtAPD = wgts.append
	for i in range(FnMe.numEdges):
		tpl = FnMe.getEdgeVertices(i)
		if(tpl in edgDict):
			idxAPD(i)
			wgtAPD(edgDict[tpl] * coef)
	FnMe.setCreaseEdges(idxs, wgts)

	if(__DEBUG): print("ApplyCreaseEdge:", idxs, wgts)

#### ========================================================================
def __MakeMesh_Vert2(pntNum, pnts, pivot, opts):
	"""頂点座標値リストを作成.LW/MayaのZ方向を吸収. result: MFloatPointArray"""
	# 遅い:list表記(Point())→FPA(list), list.append(Point())→FPA(list), より速い
	fpa = om2.MFloatPointArray().setLength(pntNum)
	if(opts.usePivot):
		offs = __FnVMul(pivot, opts.vScl)
		i = 0
		for p in pnts:	# enumer()より速い
			v = __FnVMulSub(pnts[i], opts.vScl, offs)
			fpa[i].x = v[0]
			fpa[i].y = v[1]
			fpa[i].z = v[2]
			i += 1
	else:
		i = 0
		for p in pnts:	# enumer()より速い
			v = __FnVMul(pnts[i], opts.vScl)
			fpa[i].x = v[0]
			fpa[i].y = v[1]
			fpa[i].z = v[2]
			i += 1

	if(__DEBUG): print(__gFMT_MAKE_VERT % pntNum)
	return(fpa)

#### ========================================================================
def __MakeMesh_PolPntNum_Order2(pols):
	"""perポリ頂点数/順リスト(MIntArray).
	polPntNum[pntNum,pntNum,...]. order[pntIdx,pntIdx,...]
		各ポリの頂点数の配列(MIntArray),
		各頂点idx配列を全て連結した(MIntArray)
	"""
	(arPntNums, arPolPntIdxs) = __MakeList_PolPntNum_Order_Colection(pols)

	iaPNums = om2.MIntArray(arPntNums)			# ポリ構成頂点数の配列
	iaPolPntIdxs = om2.MIntArray(arPolPntIdxs)	# 使用頂点Idxの配列
	return(iaPNums, iaPolPntIdxs)

#### ========================================================================
def __MakeMesh_Collection2(lyr, opts):
	# 係数,±Z方向適用後の頂点座標値リスト
	fpaVerts = __MakeMesh_Vert2(lyr.pntNum, lyr.pnts, lyr.piv, opts)
	# ポリ頂点数リストと反時計回り化した頂点順リスト
	(iaPols, iaOrd) = __MakeMesh_PolPntNum_Order2(lyr.pols)
	return(fpaVerts, iaPols, iaOrd)
#### ----------------
def __MakeMesh_Build2(name, fpaVerts, iaPols, iaOrd, lyr, opts):
	(dagpMesh, moMesh) = __MakeMesh_wTransforms2(name, fpaVerts, iaPols, iaOrd)
	if(opts.usePivot):
		__ApplyPivotOffsetToTransform2(dagpMesh, opts.vScl, lyr.piv)
	return(dagpMesh, moMesh)

#### ========================================================================
def __SetUVSet2(moMesh, uvs):
	if(__DEBUG): print("SetUVSet:")

	uvNames = uvs.keys()	# .[name][vmap/vmad]...
	if(len(uvNames) == 0): return(False, None)

	FnMe = om2.MFnMesh(moMesh)
	mapNames = {}	# {旧名:真名}
	for name in uvNames:
		if(name == DefaultNM.UV):
			mapNames.update({ DefaultNM.UV: DefaultNM.UV })
			# 重複作成はしない.値を設置するまでゴースト状態？
		else:
			newName = FnMe.createUVSet(name)
			mapNames.update({ name: newName })

	if(DefaultNM.UV in mapNames):	# map1があるなら優先してカレントにする
		FnMe.setCurrentUVSetName(DefaultNM.UV)
	else:	# 旧名[0]の真名を辞書から拾う
		FnMe.setCurrentUVSetName(mapNames[uvNames[0]])
		# 新規作成に際してmap1が存在するが消せない

	return(True, mapNames)

#### ========================================================================
def __AssignUV2(moMesh, polNum, u, v, vmad, name):
	(polPntNums, uvIdxs) = ([], [])
	ppnAPD = polPntNums.append
	uviEXT = uvIdxs.extend
	for i in range(polNum):
		if(i in vmad):
			ppnAPD(len(vmad[i]))
			uviEXT(vmad[i])
		else:
			ppnAPD(0)
	# assignUVsはperPolUVとして全てのポリに作用する為、
	# perPolUVが成立しないポリ[Idx]部は無効フラグ"0"で埋める必要がある

	iPolPntNums = om2.MIntArray(polPntNums)
	iuvIdxs     = om2.MIntArray(uvIdxs)
	FnMe = om2.MFnMesh(moMesh)
	FnMe.setUVs(u, v, name)		# 旧:真map名
	FnMe.assignUVs(iPolPntNums, iuvIdxs, name)

#### ========================================================================
#### ----------------
def __BlendShape_Mesh2(fpaVerts, iaPols, iaOrd, FnTrMorfGrp, morphs, isRelMorph, lyr, opts):
	FnTr = om2.MFnTransform()	# 汎用
	FnMe = om2.MFnMesh()		# 汎用

	(moTrsDict, morfDict) = ({}, {})
	for (baseName, idValDct) in sorted(morphs.items()):	# {id:pos, id:pos,...}
		# メッシュ作成, 未確定Meshへの変形と削除は不正処理される.
		# 元型create→copy→get/setPoints→deleteVertex == NG
		# 元型create→get/setPoints→deleteVertex == NG
		# PointArrayの複製を加工してcreate→deleteVertexする形にする
		fpaVt = om2.MFloatPointArray(fpaVerts)
		if(isRelMorph):		# 相対
			__BlendShape_Mesh_ModifyRelFPA(fpaVt, idValDct, opts.vScl)
		else:
			if(opts.usePivot):
				__BlendShape_Mesh_ModifyAbsPivFPA(fpaVt, idValDct, opts.vScl, lyr.piv)
			else:
				__BlendShape_Mesh_ModifyAbsFPA(fpaVt, idValDct, opts.vScl)
		(dagpMesh, moMesh) = __MakeMesh_Build2(lyr.name, fpaVt, iaPols, iaOrd, lyr, opts)

		# grpの追加/新規, {grpName: {morf:MObj},{},...}
		(grpName, morfName) = __MakeMesh_MorfsNames_AddDict(baseName, moMesh, morfDict)

		FnMe.setObject(moMesh)
		FnMe.setName(morfName + SFixNM.MF_SHAPE)	# 改名, 戻り名不要
		FnTr.setObject(dagpMesh.transform())		# 親Trs
		FnTr.setName(morfName + SFixNM.MF_TRS)		# 親Trsを改名, 戻り名不要

		# grpの追加/新規, {grpName: [MObj,MObj,...]}
		__MakeMesh_AddGroupDict(grpName, FnTr, moTrsDict)

		# mesh/Skelegon/curve混在対策, 未確定Meshの未使用頂点は一つ消せば全て消える
		if(lyr.bonNum):
			FnMe.deleteVertex(lyr.bones[0][BoneT.PNTID][0])	# 1本目の第1頂点のidx.
		elif(lyr.crvNum):
			FnMe.deleteVertex(lyr.crvs[0][1])	# 1本目の第1頂点のidx.

		FnMe.setObject(dagpMesh)	# dagpだとFix, Fixすると未使用頂点は消せない

	moGrpList = []		# morphGroup毎とlayer毎でグループ化
	for (grpName, moTrs) in moTrsDict.items():	# {grpName: [MObj,MObj,...]}
		moGrpList.append(__ObjectGrouping2(GrpVisT.SHOW, moTrs, "%s_grp" % grpName))

	prefix = ("Abs", "Rel")
	tmpName = "%sMorph_grp" % prefix[isRelMorph]
	mo = __ObjectGrouping2(GrpVisT.SHOW, moGrpList, tmpName)	# Rel/Absグループ名の保証
	FnTrMorfGrp.addChild(mo)
	FnTr.setObject(mo)
	FnTr.setName(tmpName)

	return(morfDict)

#### ========================================================================
def __BlendShape_Set2(dagpMesh, morfDict):	# パラレル
	def __ConvMo2To1(mo2):
		Fn = om2.MFnMesh(mo2)
		tmp1 = om.MSelectionList()
		tmp1.add(Fn.fullPathName())
		mo1 = om.MObject()
		tmp1.getDependNode(0, mo1)
		return(mo1)

	FnBl = oa.MFnBlendShapeDeformer()
	mpElms = om.MIntArray()

	dagp = __ConvertDagPath_2To1(dagpMesh)	####
	FnMe = om.MFnMesh(dagp)
	moMesh = FnMe.object()

	for (grpName, dic) in morfDict.items():
		bldName = grpName + SFixNM.BLSHAPE
		if(mc.objExists(bldName)):						# blendShapeが既存
			moBl = __Get_MObject_ByName([bldName])[0]
			FnBl.setObject(moBl)
			idxOffset = mc.getAttr(bldName + ".w", s=True)
		else:
			FnBl.create(moMesh)		# ノード作成
			FnBl.setName(bldName)	# BlShapeにグループ名を付ける.
			idxOffset = 0

		for (i, (morfName, mo)) in enumerate(sorted(dic.items())):	# om2のmo
			mo1 = __ConvMo2To1(mo)
			idx = i + idxOffset
			FnBl.addTarget(moMesh, idx, mo1, 1.0)			# blShapeNodeは維持される
			# .weight[]のエリアス名が表示される.初期のエリアス名はshape名.
			mp = FnBl.findPlug("weight")					# 配列のルートを取る
			FnBl.setAlias(morfName, "weight[%d]" % idx, mp)	# [N]が必要.

#### ========================================================================
def __MakeMesh_wTransforms2(name, fpaVerts, iaPols, iaOrd):
	FnTr = om2.MFnTransform()
	FnTr.create()
	FnMe = om2.MFnMesh()
	FnMe.create(fpaVerts, iaPols, iaOrd, parent=FnTr.object())	# createの戻りmoは不定？
	trName = FnTr.setName(name)		# 重複時は既定の名称変化
	FnMe.setName(trName + "Shape")
	return(FnMe.getPath(), FnMe.object())

#### ========================================================================
def __ApplyPivotOffsetToTransform2(dagp, vScl, pivot):
	"""pivotが原点となるように頂点座標を修正"""
	FnTr = om2.MFnTransform(dagp.transform())	# FnTr.setTranslationはkWorldを受けない
	FnTr.setTranslation(om2.MVector(*__FnVMul(pivot, vScl)), om2.MSpace.kTransform)

#### ========================================================================
def __BlendShape_Group2(name):
	FnTr = om2.MFnTransform()
	FnTr.create()
	FnTr.setName(name + SFixNM.MF_GRP)	# 戻り名不要
	FnTr.findPlug("v", False).setBool(False)
	return(FnTr)

#### ========================================================================
def __SetAttrToBaseMesh_withUpdateSurf2(moMesh, hasVertColor, opts):
	FnMe = om2.MFnMesh(moMesh)
	# カラーチャネル dcc/displayColorChannel
	# "None", "Ambient", "Ambient+Diffuse", "Diffuse", "Specular", "Emission"
	FnMe.findPlug("dcc", False).setString(__gMeshDispColrChans[3])
	# 両面 doubleSided
	FnMe.findPlug("ds", False).setBool(opts.isDblSide)
	# .dcol/displayColors
	if(hasVertColor != None): FnMe.displayColors = hasVertColor

	FnMe.updateSurface()

#### ========================================================================
def __MakeMesh_MorfTargetMesh2(fpaVerts, iaPols, iaOrd, lyr, opts):
	(hasRM, hasAM) = (len(lyr.relMorphs), len(lyr.absMorphs))
	(relMorfDict, absMorfDict) = ({}, {})
	if(opts.loadRM or opts.loadAM):	# モーフジオメトリ作成ならオブジェクトと対になるグループを作る
		if(__DEBUG): print("CreateMorph:")

		if(hasRM + hasAM):
			FnTrMorfGrp = __BlendShape_Group2(lyr.name)
		if(opts.loadRM and hasRM):
			relMorfDict = __BlendShape_Mesh2(fpaVerts, iaPols, iaOrd, FnTrMorfGrp, lyr.relMorphs, MorfT.REL, lyr, opts)
		if(opts.loadAM and hasAM):
			absMorfDict = __BlendShape_Mesh2(fpaVerts, iaPols, iaOrd, FnTrMorfGrp, lyr.absMorphs, MorfT.ABS, lyr, opts)
	return(hasRM, hasAM, relMorfDict, absMorfDict)

#### ========================================================================
def __MakeMesh_SetUVs2(moMesh, lyr):
	"""UVセット作成とカレントをセット"""
	(hasUV, mapNames) = __SetUVSet2(moMesh, lyr.uvs)
	if(hasUV):	# UVの適用(vmap&vmadの都合で適用漏れポリゴンが発生するかも.要検証)
		for name in lyr.uvs.keys():
			(u, v, vmad) = __MakeList_UV_Idxs(lyr.pols, lyr.uvs[name])	# 旧map名
			__AssignUV2(moMesh, len(lyr.pols), u, v, vmad, mapNames[name])	# 旧:真map名

#### ========================================================================
def _MakeMeshObject2(tags, lyr, surfs, clips, opts):
	if((not lyr.pntNum) or (not lyr.polNum)): return(tuple([None]*4))

	FMT = u"Geometry 構築 ({}: {}, Scale: {:g}, PivotOffs: {:g}, {:g}, {:g})"
	print(FMT.format(lyr.idx +1, lyr.name, opts.scl, *lyr.piv))

	if(__DEBUG): print("CreateMesh:")

	(fpaVerts, iaPols, iaOrd) = __MakeMesh_Collection2(lyr, opts)
	(dagpMesh, moMesh) = __MakeMesh_Build2(lyr.name, fpaVerts, iaPols, iaOrd, lyr, opts)

	# UVセット作成とカレントをセット
	__MakeMesh_SetUVs2(moMesh, lyr)

	# エッジウェイト適用
	if(opts.useEdge): __ApplyCreaseEdge2(moMesh, lyr)

	# perPolノーマル適用

	# 頂点カラー
	(rgbNames, rgbaNames, hasVertColor) = ([], [], None)
	if(opts.useVCol):
		# 頂点カラー適用
		rgbNames  = __ApplyVColors2(moMesh, lyr.pols, lyr.rgb,  opts.useMiVC)
		rgbaNames = __ApplyVColors2(moMesh, lyr.pols, lyr.rgba, opts.useMiVC)
		# 頂点カラー表示
		hasVertColor = True if(rgbNames or rgbaNames) else False

	# サーフェス(SG)適用 om2.dagpath からom1へ. om2(2015)にMFnSetやMFnマテリアルは無い
	(mtNames, sgNames) = __MakeSurface(tags, lyr.surfTags, __ConvertDagPath_2To1(dagpMesh), opts)
	# マテリアルの設定
	__SetSurfaceAttr(surfs, clips, mtNames, sgNames, rgbNames, rgbaNames, opts)

	# セレクションセット作成 om2.dagpath からom1へ. om2(2015)にMFnSetは無い
	if(opts.usePick): __Set_SelectionSet(__ConvertDagPath_2To1(dagpMesh), lyr.name, lyr.picks, True)
	if(opts.usePCol): __Set_SelectionSet(__ConvertDagPath_2To1(dagpMesh), lyr.name, lyr.tagColors, False)

	# レイヤがPoly/Skelegon/Curve混在の場合、頂点が未使用頂点としてmesh内に残る.
	# 削除によるindexズレ(uv等)対策をするのではなく、ジオメトリ構築後に削除する.
	# [ 1点でも削除すれば全ての未使用頂点が消え、indexが詰められる.]
	# MayaUI操作ではエッジの削除等でトポロジ変更を起こせば消える(頂点単位の明示的削除は出来ない).
	if(lyr.bonNum):
		FnMe = om2.MFnMesh(moMesh)
		FnMe.deleteVertex(lyr.bones[0][BoneT.PNTID][0])	# 1本目の第1頂点のidx.
##		if(len(lyr.weights) and opts.useWgt):	# ボーンが参照するvmap値を保持する場合.
##			__MakeWeightMap(dagpMesh.fullPathName(), lyr.bones, lyr.weights, lyr.pntNum)
	elif(lyr.crvNum):
		FnMe = om2.MFnMesh(moMesh)
		FnMe.deleteVertex(lyr.crvs[0][1])	# 1本目の第1頂点のidx.

	FnMe = om2.MFnMesh(dagpMesh)	# dagpだとFix, Fixすると未使用頂点は消せない

	# モーフ形状作成
	(hasRM, hasAM, relMorfDict, absMorfDict) = __MakeMesh_MorfTargetMesh2(fpaVerts, iaPols, iaOrd, lyr, opts)

	# ブレンドシェイプ作成 om2.dagpath からom1へ. om2(2015)にMFnBlendShapeDefomerは無い
	if(opts.useRM and opts.loadRM and hasRM): __BlendShape_Set2(dagpMesh, relMorfDict)
	if(opts.useAM and opts.loadAM and hasAM): __BlendShape_Set2(dagpMesh, absMorfDict)

	__SetAttrToBaseMesh_withUpdateSurf2(moMesh, hasVertColor, opts)

	if(__DEBUG):
		pass

	return("", "", "", "")

#### ========================================================================
def __ObjectGrouping2(isVis, arTrs, name):
	'''トランスフォームを作り,arTrs内のmoを子にする'''
	FnTrs = om2.MFnTransform()
	FnTrs.create()
	FnTrs.setName(name)
	if(isVis == GrpVisT.HIDE): FnTrs.findPlug("v", False).setBool(isVis)

	for mo in arTrs:
		FnTrs.addChild(mo)
	return(FnTrs.object())	# MObjectを返す

#### ========================================================================
def __ApplyVColors2(moMesh, pols, RGB, useMiVC):
	if(__DEBUG): print("ApplyVColor:")

	FnMe = om2.MFnMesh(moMesh)

	if(useMiVC): rgbNames = __SetVColorSet2_mrVCol(FnMe, RGB)
	else:		 rgbNames = __SetVColorSet2(FnMe, RGB)
	if(rgbNames == None): return(None)

	for name in RGB.keys():
		FnMe.setCurrentColorSetName(rgbNames[name])		# 旧:真map名
		if((ChunkT.VMAP in RGB[name]) and (ChunkT.VMAD in RGB[name])):
			__ApplyVColors2_Mixed(FnMe, RGB[name], pols)
		else:
			if(ChunkT.VMAP in RGB[name]): __ApplyVColors2_VMAP(FnMe, RGB[name])
			if(ChunkT.VMAD in RGB[name]): __ApplyVColors2_VMAD(FnMe, RGB[name])

	return(rgbNames)

#### ========================================================================
def __SetVColorSet2_mrVCol(FnMe, rgbs):
	colNames = rgbs.keys()	# .[name][vmap/vmad]...
	if(len(colNames) == 0): return(None)

	meName = FnMe.fullPathName()		# ※パス無しだと下記getAttr(s)時に余計な物も拾ってくる, miなら平気？
	meNode = meName + ".clst[%d].clsn"	# .colorSet[%d].colorName
	idxs = range(len(colNames))
	# sizeでの複数時の戻りはidxのArray[N,...,0], ※同名Meshがあると複数扱いされる
	# miは[0,...N]

	rex = re.compile("[\[\]= +-/~^|?/*]")

	mapNames = {}	# {旧名:真名}
	for (i, name) in enumerate(colNames):
		newName = FnMe.createColorSet(rex.sub("_", name), True)	# clampさせる
		mapNames.update({ name: newName })

		# ColSet名のmrVColを重複作成はしない
		mrvcName = PFixNM.VCOL + newName
		if(mc.objExists(mrvcName)):
			__CleaningVColSlot(mrvcName)		# 既存の空きslot削除
		else:
			mrvcName = mc.shadingNode(NodeT.VCOL, at=True, n= mrvcName)

		src = meNode % idxs[i]

		id = __GetVColSlotIdx(mrvcName)
		dst = mrvcName + (".s[%d]" % id)	# .cpvSets[%d]"
		mc.connectAttr(src, dst, f=True)
	return(mapNames)

#### ========================================================================
def __SetVColorSet2(FnMe, rgbs):
	colNames = rgbs.keys()	# .[name][vmap/vmad]...
	if(len(colNames) == 0): return(None)

	rex = re.compile("[\[\]= +-/~^|?/*]")

	mapNames = {}	# {旧名:真名}
	for name in colNames:
		newName = FnMe.createColorSet(rex.sub("_", name), True)	# clampさせる
		mapNames.update({ name: newName })
	return(mapNames)

#### ========================================================================
def __ApplyVColors2_Mixed(FnMe, rgb, pols):
	(arPols, arPnts) = ([], [])
	polAPD = arPols.append
	pntAPD = arPnts.append
	mcaCol = om2.MColorArray()
	mcaAPD = mcaCol.append
	for (polID, pntIDs) in enumerate(pols):
		if(polID in rgb[ChunkT.VMAD]):
			for pnt in pntIDs:
				if(pnt in rgb[ChunkT.VMAD][polID]):
					polAPD(polID)
					pntAPD(pnt)
					mcaAPD(om2.MColor(rgb[ChunkT.VMAD][polID][pnt]))
				else:
					if(pnt in rgb[ChunkT.VMAP]):
						polAPD(polID)
						pntAPD(pnt)
						mcaAPD(om2.MColor(rgb[ChunkT.VMAP][pnt]))
		else:
			for pnt in pntIDs:
				if(pnt in rgb[ChunkT.VMAP]):
					polAPD(polID)
					pntAPD(pnt)
					mcaAPD(om2.MColor(rgb[ChunkT.VMAP][pnt]))
	FnMe.setFaceVertexColors(mcaCol, om2.MIntArray(arPols), om2.MIntArray(arPnts))

#### ========================================================================
def __ApplyVColors2_VMAD(FnMe, rgb):
	(arPols, arPnts) = ([], [])
	polAPD = arPols.append
	pntAPD = arPnts.append
	mcaCol = om2.MColorArray()
	mcaAPD = mcaCol.append
	for (polIdx, dic) in rgb[ChunkT.VMAD].items():
		for (pntIdx, col) in dic.items():
			polAPD(polIdx)
			pntAPD(pntIdx)
			mcaAPD(om2.MColor(col))	# col[0~4]
	FnMe.setFaceVertexColors(mcaCol, om2.MIntArray(arPols), om2.MIntArray(arPnts))
	'''
	for (polIdx, dic) in rgb[ChunkT.VMAD].items():
		for (pntIdx, col) in dic.items():
			print(polIdx, pntIdx, col)
			mcCol = om2.MColor(col)	# col[0~4]
			FnMe.setFaceVertexColor(mcCol, polIdx, 0)#pntIdx)	# pntidx in pol BUG?
	'''

#### ========================================================================
def __ApplyVColors2_VMAP(FnMe, rgb):
	arPnts = [ idx for idx in rgb[ChunkT.VMAP].keys() ]
	mcaCol = om2.MColorArray().setLength(len(rgb[ChunkT.VMAP]))
	i = 0
	for col in rgb[ChunkT.VMAP].values():
		mcaCol[i].r = col[0]
		mcaCol[i].g = col[1]
		mcaCol[i].b = col[2]
		mcaCol[i].a = col[3]
		i += 1
	FnMe.setVertexColors(mcaCol, om2.MIntArray(arPnts))

