﻿# -*- coding: utf-8 -*-

import maya.OpenMaya as om

import FnOM1
from .. import FnCalc
from .Defines import *

#### ========================================================================
__all__ = ("MakeMesh_Collection", "MakeMesh_Build", "MakeMesh_SetUVs")

#### =======================================================================
__DEBUG = 0

#### =======================================================================
def MakeMesh_Collection(lyr, opts):
	# 係数,±Z方向適用後の頂点座標値リスト
	fpaVerts = __Make_Vert(lyr.pntNum, lyr.pnts, lyr.piv, opts)
	# ポリ頂点数リストと反時計回り化した頂点順リスト
	(newPolNum, buildedPolIdxs, miaPols, miaOrd) = __Make_PolPntNum_Order(lyr.polNum, lyr.pols)
	return(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd)

#### ------------------------------------------------------------------------
def __Make_Vert(pntNum, pnts, pivot, opts):
	u'''頂点座標値リストを作成. result: vNum, MFloatPointArray'''
	# LW/MayaのZ方向はvSclで事前に吸収.
	# set()に(,,-v) 直入れはNG. (-v)括弧で括るかv*-1表記
	# py.om1ではfp/p array, flt[][4]からの一括作成は無理
	fpa = om.MFloatPointArray(pntNum)
	fpaSET = fpa.set	# (idx, x,y,z)
	if(opts.usePivot):
		offs = FnCalc.VMul(pivot, opts.vScl)
		for (i, pnt) in enumerate(pnts):	# cnt+=1と速度差無し
			fpaSET(i, *FnCalc.VMulSub(pnt, opts.vScl, offs))
	else:
		for (i, pnt) in enumerate(pnts):
			fpaSET(i, *FnCalc.VMul(pnt, opts.vScl))

	if(__DEBUG): print(u"vert: {}".format(pntNum))
	return(fpa)

#### ------------------------------------------------------------------------
def __Make_PolPntNum_Order(polNum, pols):
	u'''perポリ頂点数/順リスト(MIntArray).
	polPntNum[pntNum,pntNum,...]. order[pntIdx,pntIdx,...]
		各ポリの頂点数の配列(MIntArray),
		各頂点idx配列を全て連結した(MIntArray)'''
	arPntNums = []			# ポリ構成頂点数の配列
	arPolPntIdxs = []		# 使用頂点Idxの配列
	pnAPD = arPntNums.append
	ppEXT = arPolPntIdxs.extend

	buildedPolIdxs = [None] * polNum
	polCnt = 0
	for (i, polIdxs) in enumerate(pols):
		n = len(polIdxs)
		if(n > 2):
			pnAPD(n)
			ppEXT(polIdxs[::-1])	# 左右逆回りのIdx
			buildedPolIdxs[i] = polCnt
			polCnt += 1

	if(__DEBUG): print(u"pol: {}\norder: {}".format(len(pols), len(arPolPntIdxs)))

	miaPntNums = FnOM1.MIntArrayFromList(arPntNums)
	miaPolPntIdxs = FnOM1.MIntArrayFromList(arPolPntIdxs)
	return(polCnt, tuple(buildedPolIdxs), miaPntNums, miaPolPntIdxs)

#### ========================================================================
#### ========================================================================
def MakeMesh_Build(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd, lyr, opts):
	(dagpMesh, moMesh) = __Make_wTransforms(lyr.name, lyr.pntNum, newPolNum, fpaVerts, miaPols, miaOrd)
	if(opts.usePivot):
		__ApplyPivotOffsetToTransform(dagpMesh, opts.vScl, lyr.piv)
	return(dagpMesh, moMesh)

#### ------------------------------------------------------------------------
def __Make_wTransforms(name, pntNum, newPolNum, fpaVerts, miaPols, miaOrd):
	FnTr = om.MFnTransform()
	FnTr.create()
	trName = FnTr.setName(name)		# 重複時は既定の名称変化,+1,2...

	FnMe = om.MFnMesh()
	FnMe.create(pntNum, newPolNum, fpaVerts, miaPols, miaOrd, FnTr.object())	# moはparent指定次第
	FnMe.setName("{}Shape".format(trName))

	dagp = om.MDagPath()
	FnMe.getPath(dagp)
	return(dagp, FnMe.object())

#### ------------------------------------------------------------------------
def __ApplyPivotOffsetToTransform(dagp, vScl, pivot):
	u'''pivotが原点となるように頂点座標を修正'''
	FnTr = om.MFnTransform(dagp.transform())	# FnTr.setTranslationはkWorldを受けない
	FnTr.setTranslation(om.MVector(*FnCalc.VMul(pivot, vScl)), om.MSpace.kTransform)

#### ========================================================================
#### ========================================================================
def MakeMesh_SetUVs(moMesh, newPolNum, buildedPolIdxs, lyr):
	u'''UVセット作成とカレントをセット'''
	(hasUV, mapNames) = __SetUVSet(moMesh, lyr.uvs)
	if(hasUV):	# UVの適用(vmap&vmadの都合で適用漏れポリゴンが発生するかも.要検証)
		for name in lyr.uvs.keys():
			(u, v, vmad) = __MakeList_UV_Idxs(lyr.pols, lyr.uvs[name])	# 旧map名
			__AssignUV(moMesh, lyr.polNum, buildedPolIdxs, u, v, vmad, mapNames[name])	# 旧:真map名

#### ------------------------------------------------------------------------
def __SetUVSet(moMesh, uvs):
	if(__DEBUG): print("SetUVSet:")

	uvNames = uvs.keys()	# .[name][vmap/vmad]...
	if(len(uvNames) == 0): return(False, None)

	FnMe = om.MFnMesh(moMesh)
	mapNames = {}	# {旧名:真名}
	for name in uvNames:
		if(name == DEFAULT_UV):
			mapNames.update({ DEFAULT_UV: DEFAULT_UV })
			# 重複作成はしない.値を設置するまでゴースト状態？
		else:
			newName = FnMe.createUVSetWithName(name)
			mapNames.update({ name: newName })

	if(DEFAULT_UV in mapNames):	# map1があるなら優先してカレントにする
		FnMe.setCurrentUVSetName(DEFAULT_UV)
	else:	# 旧名[0]の真名を辞書から拾う
		FnMe.setCurrentUVSetName(mapNames[uvNames[0]])
		# 新規作成に際してmap1が存在するが消せない

	return(True, mapNames)

#### ------------------------------------------------------------------------
def __MakeList_UV_Idxs(pols, uvs):
	(cnt, u, v, vmap) = __MakeList_UV_VMap(uvs)
	(u, v, vmad) = __MakeList_UV_VMad(pols, uvs, cnt, u, v, vmap)
	return(u, v, vmad)

#### ------------------------------------------------------------------------
def __MakeList_UV_VMap(uvs):
	u'''u/vの配列分離.pntIdxからU/V[idx]を指す{pntIdx:u/v idx}の作成'''
	u = om.MFloatArray()
	v = om.MFloatArray()
	vmap = {}
	if(not uvs.vmap):
		return(0, u, v, vmap)	# 空を返す.vmadに続く.

	uAPD = u.append
	vAPD = v.append
	UPD = vmap.update
	uvIdx = 0
	for (pntIdx, uv) in uvs.vmap.items():	# vmap{pntIdx:(u,v)},{},...
		uAPD(uv[0])				# 頂点idxからuv配列idxを指す
		vAPD(uv[1])				# pntがUVを持つ保障は無い
		UPD({ pntIdx: uvIdx })	# pntIdxとuvIdxはずれ得る
		uvIdx += 1
	return(uvIdx, u, v, vmap)

#### ------------------------------------------------------------------------
def __MakeList_UV_VMad(pols, uvs, uvCnt, u, v, vmap):
	u'''u/vの配列分離.polIdxから指す{polIdx:{pntIdx:u/v idx}...}の作成'''
	#### -----------------------------------
	def Hoge_VMap(polIdx, pntIdxs, vmap, vmad):
		polPnt = []
		APD = polPnt.append
		for pntIdx in reversed(pntIdxs):
			if(pntIdx in vmap):
				APD(vmap[pntIdx])
		if(len(pntIdxs) == len(polPnt)):		# 全点がuv/vmapを持つ
			 vmad.update({ polIdx: polPnt })	# { 旧polIdx:[uvIdx,...] }
	#### -----------------------------------
	def Hoge_VMad(polIdx, pntIdxs, uvs, u, v, vmap, vmad, uvCnt):
		polPnt = []
		ppnAPD = polPnt.append
		uAPD = u.append
		vAPD = v.append
		for pntIdx in reversed(pntIdxs):
			if(pntIdx in uvs.vmad[polIdx]):
				uv = uvs.vmad[polIdx][pntIdx]
				uAPD(uv[0])
				vAPD(uv[1])
				ppnAPD(uvCnt)
				uvCnt += 1
			else:
				if(pntIdx in vmap):
					ppnAPD(vmap[pntIdx])

		if(len(pntIdxs) == len(polPnt)):		# 全点がuv/vmapを持つ
			vmad.update({ polIdx: polPnt })		# { 旧polIdx:[uvIdx,...] ]
		return(uvCnt)
	#### -----------------------------------
	vmad={}
	polNum = len(pols)
	if(uvs.vmad):	# 不連続UV有り
		for (polIdx, pntIdxs) in enumerate(pols):
			if(len(pntIdxs) > 2):
				if(polIdx in uvs.vmad):			# polyがVMADを持つなら
					uvCnt = Hoge_VMad(polIdx, pntIdxs, uvs, u, v, vmap, vmad, uvCnt)
				else:							# polyがVMAPのみ
					Hoge_VMap(polIdx, pntIdxs, vmap, vmad)
	else:			# 連続UVのみ
		for (polIdx, pntIdxs) in enumerate(pols):
			if(len(pntIdxs) > 2):
				Hoge_VMap(polIdx, pntIdxs, vmap, vmad)
	return(u, v, vmad)

#### ------------------------------------------------------------------------
def __AssignUV(moMesh, polNum, buildedPolIdxs, u, v, vmad, name):
	polPntNums = []
	uvIdxs = []
	ppnAPD = polPntNums.append
	uviEXT = uvIdxs.extend
	for i in buildedPolIdxs:
		if(i != None):
			if(i in vmad):
				ppnAPD(len(vmad[i]))
				uviEXT(vmad[i])
			else:
				ppnAPD(0)
	# assignUVsはperPolUVとして全てのポリに作用する為、
	# perPolUVが成立しないポリ[Idx]部は無効フラグ"0"で埋める必要がある

	miaPolPntNums = FnOM1.MIntArrayFromList(polPntNums)
	miaUvIdxs = FnOM1.MIntArrayFromList(uvIdxs)
	FnMe = om.MFnMesh(moMesh)
	FnMe.setUVs(u, v, name)		# 旧:真map名
	FnMe.assignUVs(miaPolPntNums, miaUvIdxs, name)

#### ========================================================================
#### ========================================================================

